/***

   This file reduces the size of some GTK widgets by reducing unnecessary padding.
   You may make changes or comment-out text to restore the standard GTK widgets.
   Do not delete the file - it would be automatically restored.

***/


/***  disabled modifications

window *, menu, box * {
   color: #000000;
   background-color: #DDDDDD;
}

***/


button, .button {
   padding: 0px 2px 0px 2px;        /* top right bottom left */ 
}

scale {                             /* slider control */
   padding: 6px 6px;
}



